#include "localsolver.h"
#include <fstream>
#include <iostream>
#include <string.h>
#include <vector>

using namespace localsolver;
using namespace std;

class AircraftLanding {
private:
    // Data from the problem
    int nbPlanes;
    int tmp;
    vector<int> earliestTimeData;
    vector<int> targetTimeData;
    vector<int> latestTimeData;
    vector<double> earlinessCostData;
    vector<double> latenessCostData;
    vector<vector<int>> separationTimeData;

    // LocalSolver
    LocalSolver localsolver;

    // Decision variables
    LSExpression landingOrder;
    vector<LSExpression> preferredTimeVars;

    // Landing time for each plane
    LSExpression landingTime;

    // Objective
    LSExpression totalCost;

public:
    /* Read instance data */
    void readInstance(const string& fileName) {
        ifstream infile;
        infile.exceptions(ifstream::failbit | ifstream::badbit);
        infile.open(fileName.c_str());

        infile >> nbPlanes;
        infile >> tmp; // Skip freezeTime value
        earliestTimeData.resize(nbPlanes);
        targetTimeData.resize(nbPlanes);
        latestTimeData.resize(nbPlanes);
        earlinessCostData.resize(nbPlanes);
        latenessCostData.resize(nbPlanes);
        separationTimeData.resize(nbPlanes, vector<int>(nbPlanes));
        preferredTimeVars.resize(nbPlanes);

        for (int i = 0; i < nbPlanes; ++i) {
            infile >> tmp; // Skip appearanceTime values
            infile >> earliestTimeData[i];
            infile >> targetTimeData[i];
            infile >> latestTimeData[i];
            infile >> earlinessCostData[i];
            infile >> latenessCostData[i];
            for (int j = 0; j < nbPlanes; ++j) {
                infile >> separationTimeData[i][j];
            }
        }
        infile.close();
    }

    LSExpression getMinLandingTime(LSExpression p, LSExpression prev, LSModel model, LSExpression separationTime) {
        return model.iif(p > 0, prev + model.at(separationTime, landingOrder[p - 1], landingOrder[p]), 0);
    }

    void solve(int limit) {
        // Declare the optimization model
        LSModel model = localsolver.getModel();

        // A list variable: landingOrder[i] is the index of the ith plane to land
        landingOrder = model.listVar(nbPlanes);

        // All planes must be scheduled
        model.constraint(model.count(landingOrder) == nbPlanes);

        // Create LocalSolver arrays to be able to access them with "at" operators
        LSExpression targetTime = model.array(targetTimeData.begin(), targetTimeData.end());
        LSExpression latestTime = model.array(latestTimeData.begin(), latestTimeData.end());
        LSExpression earlinessCost = model.array(earlinessCostData.begin(), earlinessCostData.end());
        LSExpression latenessCost = model.array(latenessCostData.begin(), latenessCostData.end());
        LSExpression separationTime = model.array();
        for (int i = 0; i < nbPlanes; ++i) {
            LSExpression row = model.array(separationTimeData[i].begin(), separationTimeData[i].end());
            separationTime.addOperand(row);
        }

        // Int variables: preferred time for each plane
        for (int p = 0; p < nbPlanes; ++p) {
            preferredTimeVars[p] = model.intVar(earliestTimeData[p], targetTimeData[p]);
        }
        LSExpression preferredTime = model.array(preferredTimeVars.begin(), preferredTimeVars.end());

        // Landing time for each plane
        LSExpression landingTimeLambda = model.createLambdaFunction([&](LSExpression p, LSExpression prev) {
            return model.max(preferredTime[landingOrder[p]], getMinLandingTime(p, prev, model, separationTime));
        });
        landingTime = model.array(model.range(0, nbPlanes), landingTimeLambda);

        // Landing times must respect the separation time with every previous plane
        for (int p = 1; p < nbPlanes; ++p) {
            LSExpression lastSeparationEnd = model.max();
            for (int previousPlane = 0; previousPlane < p; ++previousPlane) {
                lastSeparationEnd.addOperand(landingTime[previousPlane] +
                                             model.at(separationTime, landingOrder[previousPlane], landingOrder[p]));
            }
            model.constraint(landingTime[p] >= lastSeparationEnd);
        }

        totalCost = model.sum();
        for (int p = 0; p < nbPlanes; ++p) {
            // Constraint on latest landing time
            LSExpression planeIndex = landingOrder[p];
            model.constraint(landingTime[p] <= latestTime[planeIndex]);

            // Cost for each plane
            LSExpression unitCost =
                model.iif(landingTime[p] < targetTime[planeIndex], earlinessCost[planeIndex], latenessCost[planeIndex]);
            LSExpression differenceToTargetTime = model.abs(landingTime[p] - targetTime[planeIndex]);
            totalCost.addOperand(unitCost * differenceToTargetTime);
        }

        // Minimize the total cost
        model.minimize(totalCost);

        model.close();

        // Parameterize the solver
        localsolver.getParam().setTimeLimit(limit);

        localsolver.solve();
    }

    /* Write the solution in a file with the following format:
     * - 1st line: value of the objective;
     * - 2nd line: for each position p, index of plane at position p. */
    void writeSolution(const string& fileName) {
        ofstream outfile;
        outfile.exceptions(ofstream::failbit | ofstream::badbit);
        outfile.open(fileName.c_str());
        outfile << totalCost.getDoubleValue() << endl;
        LSCollection landingOrderCollection = landingOrder.getCollectionValue();
        for (int i = 0; i < nbPlanes; ++i) {
            outfile << landingOrderCollection[i] << " ";
        }
        outfile << endl;
    }
};

int main(int argc, char** argv) {
    if (argc < 2) {
        cerr << "Usage: aircraft_landing inputFile [outputFile] [timeLimit]" << endl;
        return 1;
    }
    const char* instanceFile = argv[1];
    const char* solFile = argc > 2 ? argv[2] : NULL;
    const char* strTimeLimit = argc > 3 ? argv[3] : "20";
    try {
        AircraftLanding model;
        model.readInstance(instanceFile);
        model.solve(atoi(strTimeLimit));
        if (solFile != NULL)
            model.writeSolution(solFile);
        return 0;
    } catch (const exception& e) {
        cerr << "An error occurred: " << e.what() << endl;
        return 1;
    }
}
