import localsolver
import sys


def read_elem(filename):
    with open(filename) as f:
        return [str(elem) for elem in f.read().split()]


#
# Read instance data
#
def read_instance(instance_file):
    file_it = iter(read_elem(instance_file))
    nb_planes = int(next(file_it))
    next(file_it)  # Skip freezeTime value
    earliest_time_data = []
    target_time_data = []
    latest_time_data = []
    earliness_cost_data = []
    tardiness_cost_data = []
    separation_time_data = []

    for p in range(nb_planes):
        next(file_it)  # Skip appearanceTime values
        earliest_time_data.append(int(next(file_it)))
        target_time_data.append(int(next(file_it)))
        latest_time_data.append(int(next(file_it)))
        earliness_cost_data.append(float(next(file_it)))
        tardiness_cost_data.append(float(next(file_it)))
        separation_time_data.append([None] * nb_planes)

        for pp in range(nb_planes):
            separation_time_data[p][pp] = int(next(file_it))

    return nb_planes, earliest_time_data, target_time_data, latest_time_data, \
        earliness_cost_data, tardiness_cost_data, separation_time_data


def get_min_landing_time(p, prev, model, separation_time, landing_order):
    return model.iif(
        p > 0,
        prev + model.at(separation_time, landing_order[p - 1], landing_order[p]),
        0)


def main(instance_file, output_file, time_limit):
    nb_planes, earliest_time_data, target_time_data, latest_time_data, \
        earliness_cost_data, tardiness_cost_data, separation_time_data = \
        read_instance(instance_file)

    with localsolver.LocalSolver() as ls:
        #
        # Declare the optimization model
        #
        model = ls.model

        # A list variable: landingOrder[i] is the index of the ith plane to land
        landing_order = model.list(nb_planes)

        # All planes must be scheduled
        model.constraint(model.count(landing_order) == nb_planes)

        # Create LocalSolver arrays to be able to access them with an "at" operator
        target_time = model.array(target_time_data)
        latest_time = model.array(latest_time_data)
        earliness_cost = model.array(earliness_cost_data)
        tardiness_cost = model.array(tardiness_cost_data)
        separation_time = model.array(separation_time_data)

        # Int variable: preferred landing time for each plane
        preferred_time_vars = [model.int(earliest_time_data[p], target_time_data[p])
                               for p in range(nb_planes)]
        preferred_time = model.array(preferred_time_vars)

        # Landing time for each plane
        landing_time_lambda = model.lambda_function(
            lambda p, prev:
                model.max(
                    preferred_time[landing_order[p]],
                    get_min_landing_time(p, prev, model, separation_time, landing_order)))
        landing_time = model.array(model.range(0, nb_planes), landing_time_lambda)

        # Landing times must respect the separation time with every previous plane
        for p in range(1, nb_planes):
            last_separation_end = [
                landing_time[previous_plane]
                + model.at(
                    separation_time,
                    landing_order[previous_plane],
                    landing_order[p])
                for previous_plane in range(p)]
            model.constraint(landing_time[p] >= model.max(last_separation_end))

        total_cost = model.sum()
        for p in range(nb_planes):
            plane_index = landing_order[p]

            # Constraint on latest landing time
            model.constraint(landing_time[p] <= latest_time[plane_index])

            # Cost for each plane
            difference_to_target_time = abs(landing_time[p] - target_time[plane_index])
            unit_cost = model.iif(
                landing_time[p] < target_time[plane_index],
                earliness_cost[plane_index],
                tardiness_cost[plane_index])
            total_cost.add_operand(unit_cost * difference_to_target_time)

        # Minimize the total cost
        model.minimize(total_cost)

        model.close()

        # Parameterize the solver
        ls.param.time_limit = time_limit

        ls.solve()

        #
        # Write the solution in a file with the following format:
        # - 1st line: value of the objective;
        # - 2nd line: for each position p, index of plane at position p.
        #
        if output_file is not None:
            with open(output_file, 'w') as f:
                f.write("%d\n" % total_cost.value)
                for p in landing_order.value:
                    f.write("%d " % p)
                f.write("\n")


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print("Usage: python aircraft_landing.py instance_file [output_file] [time_limit]")
        sys.exit(1)

    instance_file = sys.argv[1]
    output_file = sys.argv[2] if len(sys.argv) >= 3 else None
    time_limit = int(sys.argv[3]) if len(sys.argv) >= 4 else 20
    main(instance_file, output_file, time_limit)
